# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
"""
An order preserving multidict implementation
"""
from collections.abc import Mapping
from collections.abc import KeysView
from collections.abc import ItemsView
from collections.abc import ValuesView
from collections.abc import MutableMapping
from itertools import repeat
import typing as t
from typing import Any
from typing import Dict


if t.TYPE_CHECKING:
    from typeshed import SupportsKeysAndGetItem  # type: ignore
else:
    SupportsKeysAndGetItem = Mapping

SupportsKeysAndGetItemOrIterable = t.Union[
    SupportsKeysAndGetItem, t.Iterable[t.Tuple[Any, Any]]
]

_NO_DEFAULT = object()


class MultiDict(MutableMapping):
    """
    Dictionary-like object that supports multiple values per key. Insertion
    order is preserved.

    Synopsis:

        >>> from fresco.multidict import MultiDict
        >>> d = MultiDict([('a', 1), ('a', 2), ('b', 3)])
        >>> d['a']
        1
        >>> d['b']
        3
        >>> d.getlist('a')
        [1, 2]
        >>> d.getlist('b')
        [3]
        >>> list(d.allitems())
        [('a', 1), ('a', 2), ('b', 3)]

    """

    _dict: Dict[Any, Any]
    _order: t.List[Any]
    __slots__ = ("_dict", "_order")

    setdefault = MutableMapping.setdefault

    def __init__(self, other: SupportsKeysAndGetItemOrIterable = tuple(), **kwargs):
        """
        MultiDicts can be constructed in the following ways:

            1. From a sequence of ``(key, value)`` pairs:

                >>> from fresco.multidict import MultiDict
                >>> MultiDict([('a', 1), ('a', 2)])  # doctest: +ELLIPSIS
                MultiDict(...)

            2. Initialized from another MultiDict:

                >>> from fresco.multidict import MultiDict
                >>> d = MultiDict([('a', 1), ('a', 2)])
                >>> MultiDict(d)  # doctest: +ELLIPSIS
                MultiDict(...)

            3. Initialized from a regular dict:

                >>> from fresco.multidict import MultiDict
                >>> MultiDict({'a': 1})  # doctest: +ELLIPSIS
                MultiDict(...)

            4. From keyword arguments:

                >>> from fresco.multidict import MultiDict
                >>> MultiDict(a=1)  # doctest: +ELLIPSIS
                MultiDict(...)

        """
        self._order = []
        self._dict = {}
        self._update(other, True, kwargs)

    def __getitem__(self, key):
        """
        Return the first item associated with ``key``:

            >>> d = MultiDict([('a', 1), ('a', 2)])
            >>> d['a']
            1
        """
        try:
            return self._dict[key][0]
        except IndexError:
            raise KeyError(key)

    def __setitem__(self, key, value):
        """
        Set the items associated with key to a single item, ``value``.

            >>> d = MultiDict()
            >>> d['b'] = 3
            >>> d
            MultiDict([('b', 3)])
        """
        _order = [(k, v) for k, v in self._order if k != key]  # type: ignore
        _order.append((key, value))
        self._order = _order
        self._dict[key] = [value]

    def __delitem__(self, key):
        """
        Delete all values associated with ``key``
        """
        del self._dict[key]
        self._order = [(k, v) for (k, v) in self._order if k != key]

    def __iter__(self):
        """
        Return an iterator over all keys
        """
        return iter(self._dict)

    def get(self, key, default=None):
        """
        Return the first available value for key ``key``, or ``default`` if no
        such value exists:

            >>> d = MultiDict([('a', 1), ('a', 2)])
            >>> print(d.get('a'))
            1
            >>> print(d.get('b'))
            None
        """
        return self._dict.get(key, [default])[0]

    def getlist(self, key):
        """
        Return a list of all values for key ``key``.
        """
        return self._dict.get(key, [])

    def copy(self) -> "MultiDict":
        """\
        Return a shallow copy of the dictionary:

            >>> d = MultiDict([('a', 1), ('a', 2), ('b', 3)])
            >>> copy = d.copy()
            >>> copy
            MultiDict([('a', 1), ('a', 2), ('b', 3)])
            >>> copy is d
            False
        """
        return self.__class__(self)

    @classmethod
    def fromkeys(cls, seq, value=None) -> "MultiDict":
        """\
        Create a new MultiDict with keys from seq and values set to value.

        Example:

            >>> MultiDict.fromkeys(['a', 'b'])
            MultiDict([('a', None), ('b', None)])

        Keys can be repeated:

            >>> d = MultiDict.fromkeys(['a', 'b', 'a'])
            >>> d.getlist('a')
            [None, None]
            >>> d.getlist('b')
            [None]

        """
        return cls(zip(seq, repeat(value)))

    def items(self) -> ItemsView:
        """
        Return a list of ``(key, value)`` tuples. Only the first ``(key,
        value)`` is returned where keys have multiple values:

            >>> d = MultiDict([('a', 1), ('a', 2), ('b', 3)])
            >>> list(d.items())
            [('a', 1), ('b', 3)]
        """
        return {k: v[0] for k, v in self._dict.items()}.items()

    def listitems(self) -> ItemsView:
        """\
        Like ``items``, but returns lists of values:

            >>> d = MultiDict([('a', 1), ('a', 2), ('b', 3)])
            >>> list(d.listitems())
            [('a', [1, 2]), ('b', [3])]
        """
        return self._dict.items()

    def allitems(self) -> t.Iterable[tuple[t.Any, t.Any]]:
        """\
        Return ``(key, value)`` pairs for each item in the MultiDict.
        Items with multiple keys will have multiple key-value pairs returned:

            >>> from fresco.multidict import MultiDict
            >>> d = MultiDict([('a', 1), ('a', 2), ('b', 3)])
            >>> list(d.allitems())
            [('a', 1), ('a', 2), ('b', 3)]
        """
        return iter(self._order)

    def keys(self) -> KeysView:
        """\
        Return dictionary keys. Each key is returned only once, even if
        multiple values are present.

            >>> from fresco.multidict import MultiDict
            >>> d = MultiDict([('a', 1), ('a', 2), ('b', 3)])
            >>> list(d.keys())
            ['a', 'b']
        """
        return self._dict.keys()

    def values(self) -> ValuesView:
        """
        Return values from the dictionary. Where keys have multiple values,
        only the first is returned:

            >>> from fresco.multidict import MultiDict
            >>> d = MultiDict([('a', 1), ('a', 2), ('b', 3)])
            >>> list(d.values())
            [1, 3]
        """
        return {k: v[0] for k, v in self._dict.items()}.values()

    def listvalues(self) -> ValuesView:
        """\
        Return an iterator over lists of values. Each item will be the list of
        values associated with a single key.

        Example usage:

            >>> from fresco.multidict import MultiDict
            >>> d = MultiDict([('a', 1), ('a', 2), ('b', 3)])
            >>> list(d.listvalues())
            [[1, 2], [3]]
        """
        return self._dict.values()

    def pop(self, key, default=_NO_DEFAULT) -> t.Any:
        """
        Dictionary ``pop`` method. Return and remove the value associated with
        ``key``. If more than one value is associated with ``key``, only the
        first is returned.

        Example usage:

            >>> d = MultiDict([('a', 1), ('a', 2), ('b', 3)])
            >>> d.pop('a')
            1
            >>> d
            MultiDict([('a', 2), ('b', 3)])
            >>> d.pop('a')
            2
            >>> d
            MultiDict([('b', 3)])
            >>> d.pop('a')
            Traceback (most recent call last):
                ...
            KeyError: 'a'
        """
        try:
            value = self._dict[key].pop(0)
        except (KeyError, IndexError):
            if default is _NO_DEFAULT:
                raise KeyError(key)
            return default
        self._order.remove((key, value))
        return value

    def popitem(self) -> t.Tuple[Any, Any]:
        """
        Return and remove a ``(key, value)`` pair from the dictionary.

        The item popped is always the most recently added key and the first
        value associated with it:

            >>> d = MultiDict([('a', 1), ('a', 2), ('b', 3)])
            >>> d.popitem()
            ('b', 3)
            >>> d.popitem()
            ('a', 1)
            >>> d.popitem()
            ('a', 2)
            >>> d.popitem() #doctest: +ELLIPSIS
            Traceback (most recent call last):
                ...
            KeyError: 'popitem(): dictionary is empty'
        """
        try:
            key = self._order[-1][0]
        except IndexError:
            raise KeyError("popitem(): dictionary is empty")
        return key, self.pop(key)

    def update(
        self, other: SupportsKeysAndGetItemOrIterable = tuple(), /, **kwargs: Any
    ) -> None:
        r"""
        Update the MultiDict from another MultiDict, regular dictionary or a
        iterable of ``(key, value)`` pairs. New keys overwrite old keys - use
        :meth:`extend` if you want new keys to be added to old keys.

        Updating from another MultiDict:

            >>> d = MultiDict([('name', 'eric'), ('job', 'lumberjack')])
            >>> d.update(MultiDict([('mood', 'okay')]))
            >>> d
            MultiDict([('name', 'eric'), ('job', 'lumberjack'), ('mood', 'okay')])

        from a dictionary:

            >>> d = MultiDict([('name', 'eric'), ('hobby', 'shopping')])
            >>> d.update({'hobby': 'pressing wild flowers'})
            >>> d
            MultiDict([('name', 'eric'), ('hobby', 'pressing wild flowers')])

        an iterable of ``(key, value)`` pairs:

            >>> d = MultiDict([('name', 'eric'), ('occupation', 'lumberjack')])
            >>> d.update([('hobby', 'shopping'),
            ...           ('hobby', 'pressing wild flowers')])
            >>> d
            MultiDict([('name', 'eric'), ...])

        or keyword arguments:

                >>> d = MultiDict([('name', 'eric'),\
                ...                ('occupation', 'lumberjack')])
                >>> d.update(mood='okay')

        """
        self._update(other, True, kwargs)

    def extend(self, other: SupportsKeysAndGetItemOrIterable = tuple(), **kwargs):
        """
        Extend the MultiDict with another MultiDict, regular dictionary or a
        iterable of ``(key, value)`` pairs. This is similar to :meth:`update`
        except that new keys are added to old keys.
        """
        return self._update(other, False, kwargs)

    def _update(
        self,
        other: SupportsKeysAndGetItemOrIterable,
        replace: bool,
        extra_kwargs: Dict[Any, Any],
    ):
        """
        Update the MultiDict from another object and optionally kwargs.

        :param other: the other MultiDict, dict, or iterable (first positional
                      arg)
        :param replace: if ``True``, entries will replace rather than extend
                        existing entries (second positional arg)
        """
        if isinstance(other, self.__class__):
            items = list(other.allitems())
        elif isinstance(other, Mapping):
            items = list(other.items())
        else:
            items = list(other)

        if extra_kwargs:
            items += list(extra_kwargs.items())

        if replace:
            replaced = {k for k, v in items if k in self._dict}
            self._order = [(k, v) for (k, v) in self._order if k not in replaced]
            for key in replaced:
                self._dict[key] = []

        setdefault = self._dict.setdefault
        append_order = self._order.append
        for k, v in items:
            setdefault(k, []).append(v)
            append_order((k, v))

    def __repr__(self):
        """
        ``__repr__`` representation of object
        """
        return "%s(%r)" % (self.__class__.__name__, list(self.allitems()))

    def __str__(self):
        """
        ``__str__`` representation of object
        """
        return repr(self)

    def __len__(self):
        """
        Return the total number of keys stored.
        """
        return len(self._dict)

    def __eq__(self, other):
        return isinstance(other, MultiDict) and self._order == other._order

    def __ne__(self, other):
        return not (self == other)

    def clear(self):
        self._order = []
        self._dict = {}



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
"""
The :class:`Response` class models the response from your application to a
single request.
"""
from collections.abc import Iterable
from datetime import datetime
from itertools import chain
from typing import Callable
from typing import List
from typing import Tuple
from typing import Set
import typing as t
import re
import json as stdlib_json

from fresco.util.wsgi import StartResponseWrapper
from fresco.util.wsgi import ClosingIterator
from fresco.util.urls import is_safe_url
from fresco.cookie import Cookie
import fresco

__all__ = [
    "STATUS_CONTINUE",
    "STATUS_SWITCHING_PROTOCOLS",
    "STATUS_PROCESSING",
    "STATUS_OK",
    "STATUS_CREATED",
    "STATUS_ACCEPTED",
    "STATUS_NON_AUTHORITATIVE_INFORMATION",
    "STATUS_NO_CONTENT",
    "STATUS_RESET_CONTENT",
    "STATUS_PARTIAL_CONTENT",
    "STATUS_MULTI_STATUS",
    "STATUS_ALREADY_REPORTED",
    "STATUS_IM_USED",
    "STATUS_MULTIPLE_CHOICES",
    "STATUS_MOVED_PERMANENTLY",
    "STATUS_FOUND",
    "STATUS_SEE_OTHER",
    "STATUS_NOT_MODIFIED",
    "STATUS_USE_PROXY",
    "STATUS_TEMPORARY_REDIRECT",
    "STATUS_PERMANENT_REDIRECT",
    "STATUS_BAD_REQUEST",
    "STATUS_UNAUTHORIZED",
    "STATUS_PAYMENT_REQUIRED",
    "STATUS_FORBIDDEN",
    "STATUS_NOT_FOUND",
    "STATUS_METHOD_NOT_ALLOWED",
    "STATUS_NOT_ACCEPTABLE",
    "STATUS_PROXY_AUTHENTICATION_REQUIRED",
    "STATUS_REQUEST_TIMEOUT",
    "STATUS_CONFLICT",
    "STATUS_GONE",
    "STATUS_LENGTH_REQUIRED",
    "STATUS_PRECONDITION_FAILED",
    "STATUS_PAYLOAD_TOO_LARGE",
    "STATUS_URI_TOO_LONG",
    "STATUS_UNSUPPORTED_MEDIA_TYPE",
    "STATUS_RANGE_NOT_SATISFIABLE",
    "STATUS_EXPECTATION_FAILED",
    "STATUS_UNPROCESSABLE_ENTITY",
    "STATUS_LOCKED",
    "STATUS_FAILED_DEPENDENCY",
    "STATUS_UPGRADE_REQUIRED",
    "STATUS_PRECONDITION_REQUIRED",
    "STATUS_TOO_MANY_REQUESTS",
    "STATUS_REQUEST_HEADER_FIELDS_TOO_LARGE",
    "STATUS_INTERNAL_SERVER_ERROR",
    "STATUS_NOT_IMPLEMENTED",
    "STATUS_BAD_GATEWAY",
    "STATUS_SERVICE_UNAVAILABLE",
    "STATUS_GATEWAY_TIMEOUT",
    "STATUS_HTTP_VERSION_NOT_SUPPORTED",
    "STATUS_VARIANT_ALSO_NEGOTIATES",
    "STATUS_INSUFFICIENT_STORAGE",
    "STATUS_LOOP_DETECTED",
    "STATUS_NOT_EXTENDED",
    "STATUS_NETWORK_AUTHENTICATION_REQUIRED",
    "Response",
]


#: HTTP/1.1 status codes as listed in http://www.ietf.org/rfc/rfc2616.txt
HTTP_STATUS_CODES = {
    100: "Continue",
    101: "Switching Protocols",
    102: "Processing",
    200: "OK",
    201: "Created",
    202: "Accepted",
    203: "Non-Authoritative Information",
    204: "No Content",
    205: "Reset Content",
    206: "Partial Content",
    207: "Multi-Status",
    208: "Already Reported",
    226: "IM Used",
    300: "Multiple Choices",
    301: "Moved Permanently",
    302: "Found",
    303: "See Other",
    304: "Not Modified",
    305: "Use Proxy",
    307: "Temporary Redirect",
    308: "Permanent Redirect",
    400: "Bad Request",
    401: "Unauthorized",
    402: "Payment Required",
    403: "Forbidden",
    404: "Not Found",
    405: "Method Not Allowed",
    406: "Not Acceptable",
    407: "Proxy Authentication Required",
    408: "Request Timeout",
    409: "Conflict",
    410: "Gone",
    411: "Length Required",
    412: "Precondition Failed",
    413: "Payload Too Large",
    414: "URI Too Long",
    415: "Unsupported Media Type",
    416: "Range Not Satisfiable",
    417: "Expectation Failed",
    422: "Unprocessable Entity",
    423: "Locked",
    424: "Failed Dependency",
    426: "Upgrade Required",
    428: "Precondition Required",
    429: "Too Many Requests",
    431: "Request Header Fields Too Large",
    500: "Internal Server Error",
    501: "Not Implemented",
    502: "Bad Gateway",
    503: "Service Unavailable",
    504: "Gateway Timeout",
    505: "HTTP Version Not Supported",
    506: "Variant Also Negotiates",
    507: "Insufficient Storage",
    508: "Loop Detected",
    510: "Not Extended",
    511: "Network Authentication Required",
}

# Symbolic names for the HTTP status codes
STATUS_CONTINUE = 100
STATUS_SWITCHING_PROTOCOLS = 101
STATUS_PROCESSING = 102
STATUS_OK = 200
STATUS_CREATED = 201
STATUS_ACCEPTED = 202
STATUS_NON_AUTHORITATIVE_INFORMATION = 203
STATUS_NO_CONTENT = 204
STATUS_RESET_CONTENT = 205
STATUS_PARTIAL_CONTENT = 206
STATUS_MULTI_STATUS = 207
STATUS_ALREADY_REPORTED = 208
STATUS_IM_USED = 226
STATUS_MULTIPLE_CHOICES = 300
STATUS_MOVED_PERMANENTLY = 301
STATUS_FOUND = 302
STATUS_SEE_OTHER = 303
STATUS_NOT_MODIFIED = 304
STATUS_USE_PROXY = 305
STATUS_TEMPORARY_REDIRECT = 307
STATUS_PERMANENT_REDIRECT = 308
STATUS_BAD_REQUEST = 400
STATUS_UNAUTHORIZED = 401
STATUS_PAYMENT_REQUIRED = 402
STATUS_FORBIDDEN = 403
STATUS_NOT_FOUND = 404
STATUS_METHOD_NOT_ALLOWED = 405
STATUS_NOT_ACCEPTABLE = 406
STATUS_PROXY_AUTHENTICATION_REQUIRED = 407
STATUS_REQUEST_TIMEOUT = 408
STATUS_CONFLICT = 409
STATUS_GONE = 410
STATUS_LENGTH_REQUIRED = 411
STATUS_PRECONDITION_FAILED = 412
STATUS_PAYLOAD_TOO_LARGE = 413
STATUS_URI_TOO_LONG = 414
STATUS_UNSUPPORTED_MEDIA_TYPE = 415
STATUS_RANGE_NOT_SATISFIABLE = 416
STATUS_EXPECTATION_FAILED = 417
STATUS_UNPROCESSABLE_ENTITY = 422
STATUS_LOCKED = 423
STATUS_FAILED_DEPENDENCY = 424
STATUS_UPGRADE_REQUIRED = 426
STATUS_PRECONDITION_REQUIRED = 428
STATUS_TOO_MANY_REQUESTS = 429
STATUS_REQUEST_HEADER_FIELDS_TOO_LARGE = 431
STATUS_INTERNAL_SERVER_ERROR = 500
STATUS_NOT_IMPLEMENTED = 501
STATUS_BAD_GATEWAY = 502
STATUS_SERVICE_UNAVAILABLE = 503
STATUS_GATEWAY_TIMEOUT = 504
STATUS_HTTP_VERSION_NOT_SUPPORTED = 505
STATUS_VARIANT_ALSO_NEGOTIATES = 506
STATUS_INSUFFICIENT_STORAGE = 507
STATUS_LOOP_DETECTED = 508
STATUS_NOT_EXTENDED = 510
STATUS_NETWORK_AUTHENTICATION_REQUIRED = 511

#: Mapping from python symbolic names to HTTP headers to ensure headers are
#: emitted with correct capitalization
HEADER_NAMES = {
    "accept_ranges": "Accept-Ranges",
    "age": "Age",
    "allow": "Allow",
    "cache_control": "Cache-Control",
    "connection": "Connection",
    "content_encoding": "Content-Encoding",
    "content_language": "Content-Language",
    "content_length": "Content-Length",
    "content_location": "Content-Location",
    "content_md5": "Content-MD5",
    "content_disposition": "Content-Disposition",
    "content_range": "Content-Range",
    "content_type": "Content-Type",
    "date": "Date",
    "etag": "ETag",
    "expires": "Expires",
    "last_modified": "Last-Modified",
    "link": "Link",
    "location": "Location",
    "p3p": "P3P",
    "pragma": "Pragma",
    "proxy_authenticate": "Proxy-Authenticate",
    "refresh": "Refresh",
    "retry_after": "Retry-After",
    "server": "Server",
    "set_cookie": "Set-Cookie",
    "strict_transport_security": "Strict-Transport-Security",
    "trailer": "Trailer",
    "transfer_encoding": "Transfer-Encoding",
    "vary": "Vary",
    "via": "Via",
    "warning": "Warning",
    "www_authenticate": "WWW-Authenticate",
    "x_frame_options": "X-Frame-Options",
    "x_content_type_options": "X-Content-Type-Options",
    "x_forwarded_proto": "X-Forwarded-Proto",
    "front_end_https": "Front-End-Https",
    "x_powered_by": "X-Powered-By",
    "x_ua_compatible": "X-UA-Compatible",
}

default_charset = "UTF-8"


def encoder(stream, charset):
    """
    Encode a response iterator using the given character set.
    """
    if charset is None:
        charset = default_charset

    for chunk in stream:
        if not isinstance(chunk, bytes):
            yield chunk.encode(charset)
        else:
            yield chunk


def make_header_name(name):
    """
    Return a formatted header name from a python idenfier.

    Example usage::

        >>> make_header_name('content_type')
        'Content-Type'
    """
    try:
        return HEADER_NAMES[name]
    except KeyError:
        return name.replace("_", "-").title()


def make_headers(
    header_list, header_dict, make_header_name=make_header_name, chain=chain
):
    """
    Return a list of header (name, value) tuples from the combination of
    the header_list and header_dict.

    Synopsis::

        >>> make_headers(
        ...     [('Content-Type', 'text/html')],
        ...     {'content_length' : 54}
        ... )
        [('Content-Type', 'text/html'), ('Content-Length', '54')]

        >>> make_headers(
        ...     [('Content-Type', 'text/html')],
        ...     {'x_foo' : ['a1', 'b2']}
        ... )
        [('Content-Type', 'text/html'), ('X-Foo', 'a1'), ('X-Foo', 'b2')]

    """
    hs: List[Tuple[str, str]] = []
    addheader = hs.append
    for h, val in chain(header_list, header_dict.items()):
        if val is None:
            continue
        h = make_header_name(h)
        if isinstance(val, list):
            for item in val:
                addheader((h, str(item)))
        else:
            addheader((h, str(val)))
    return hs


class Response(object):
    """
    Model an HTTP response
    """

    default_content_type = "text/html; charset=UTF-8"
    content_iterator: Iterable[bytes]

    def __init__(
        self,
        content=None,
        status=None,
        headers=None,
        onclose=None,
        _nocontent=[],
        passthrough=False,
        content_iterator=None,
        make_headers=make_headers,
        **kwargs,
    ):
        """
        Create a new Response object, modelling the HTTP status, headers and
        content of an HTTP response. Response instances are valid WSGI
        applications.

        :param content: The response content as an iterable object
        :param status: The HTTP status line, eg ``200 OK`` or ``404 Not Found``
        :param headers: A list of HTTP headers
        :param passthrough: If True, use the content iterator unmodified.
                            Default behaviour is to wrap the content iterator
                            with :func:`encoder` to encode unicode strings
                            before output.
        :param kwargs: Arbitrary headers, provided as keyword arguments.
                       Underscores will be replaced with hyphens (eg
                       ``content_length`` becomes ``Content-Length``).

        Example usage::

            >>> # Construct a response
            >>> r = Response(
            ...     content=['hello world'],
            ...     status='200 OK',
            ...     headers=[('Content-Type', 'text/plain')]
            ... )
            >>>

        Changing headers or content::

            >>> r = r.add_header('X-Header', 'hello!')
            >>> r = r.replace(content=['whoa nelly!'],
            ...               content_type='text/html')

        """
        if content is None:
            content = _nocontent
            if status is None:
                status = "204 No Content"

        if status is None:
            self.status = "200 OK"
        else:
            try:
                self.status = f"{status} {HTTP_STATUS_CODES[status]}"  # type: ignore
            except KeyError:
                self.status = str(status)

        if onclose is None:
            self.onclose: List[Callable] = []
        elif callable(onclose):
            self.onclose = [onclose]
        else:
            self.onclose = list(onclose)

        if headers is None:
            headers = []
        if headers or kwargs:
            headers = make_headers(headers, kwargs)

        # Ensure a content-type header is set if a content iterator has been
        # provided
        if content is not _nocontent:
            if not any(k == "Content-Type" for k, v in headers):
                headers = [("Content-Type", self.default_content_type)] + headers

        self.headers = headers

        # Now we've dealt with the headers (including Content-Type,
        # which the charset property will look for when deciding how to
        # encode strings) we can pull the content into an iterable.
        # We optimize for the common cases (byte string, unicode string,
        # list of byte strings, list of unicode strings), otherwise
        # we wrap the iterator in :func:`encoder`, which iterates the content
        # and tries to encode each item as a byte string. Note that this
        # code assumes that lists are homogeneous.
        self.content = content
        if content_iterator:
            self.content_iterator = content_iterator
        elif passthrough or content is _nocontent:
            self.content_iterator = content
        else:
            content_islist = isinstance(content, list)
            if content_islist:
                if len(content) == 0:
                    self.content_iterator = content
                elif isinstance(content[0], bytes):
                    self.content_iterator = content
                elif isinstance(content[0], str):
                    charset = self.charset
                    self.content_iterator = (c.encode(charset) for c in content)

            elif isinstance(content, bytes):
                self.content_iterator = [content]
            elif isinstance(content, str):
                self.content_iterator = [content.encode(self.charset)]
            else:
                self.content_iterator = encoder(content, self.charset)

    def __call__(self, environ, start_response, exc_info=None) -> Iterable[bytes]:
        """
        WSGI callable. Calls ``start_response`` with assigned headers and
        returns an iterator over ``content``.
        """
        start_response(self.status, self.headers, exc_info)
        result = self.content_iterator
        if self.onclose:
            result = ClosingIterator[bytes](result, *self.onclose)
        return result

    def add_onclose(self, *funcs):
        """
        Add functions to be called as part of the response iterator's ``close``
        method.
        """
        return self.__class__(
            self.content,
            self.status,
            self.headers,
            self.onclose + list(funcs),
            content_iterator=self.content_iterator,
        )

    @classmethod
    def from_wsgi(cls, wsgi_callable, environ, start_response):
        """
        Return a ``Response`` object constructed from the result of calling
        ``wsgi_callable`` with the given ``environ`` and ``start_response``
        arguments.
        """
        responder = StartResponseWrapper(start_response)
        content = wsgi_callable(environ, responder)
        close = getattr(content, "close", None)
        if responder.buf.tell():
            content = chain(content, [responder.buf.getvalue()])

        if not responder.called:
            # Iterator has not called start_response yet. Call next(content)
            # to force the application to call start_response
            try:
                chunk = next(content)
            except StopIteration:
                pass
            except Exception:
                if close is not None:
                    close()
                raise
            else:
                content = chain([chunk], content)
        return cls(
            content,
            responder.status,
            headers=responder.headers,
            passthrough=True,
            onclose=[close] if close is not None else [],
        )

    def get_headers(self, name):
        """
        Return the list of headers set with the given name.

        Synopsis::

            >>> r = Response(set_cookie = ['cookie1', 'cookie2'])
            >>> r.get_headers('set-cookie')
            ['cookie1', 'cookie2']

        """
        return [
            value for header, value in self.headers if header.lower() == name.lower()
        ]

    def get_header(self, name, default=""):
        """
        Return the concatenated values of the named header(s) or ``default`` if
        the header has not been set.

        As specified in RFC2616 (section 4.2), multiple headers will be
        combined using a single comma.

        Example usage::

            >>> r = Response(set_cookie = ['cookie1', 'cookie2'])
            >>> r.get_header('set-cookie')
            'cookie1,cookie2'
        """
        headers = self.get_headers(name)
        if not headers:
            return default
        return ",".join(headers)

    @property
    def status_code(self):
        """
        Return the numeric status code for the response as an integer::

            >>> Response(status='404 Not Found').status_code
            404
            >>> Response(status=200).status_code
            200
        """
        return int(self.status.split(" ", 1)[0])

    @property
    def content_type(self):
        """
        Return the value of the ``Content-Type`` header if set, otherwise
        ``None``.
        """
        for key, val in self.headers:
            if key.lower() == "content-type":
                return val
        return None

    def add_header(self, name, value, make_header_name=make_header_name):
        """
        Return a new response object with the given additional header.

        Synopsis::

            >>> r = Response(content_type='text/plain')
            >>> r.headers
            [('Content-Type', 'text/plain')]
            >>> r.add_header('Cache-Control', 'no-cache').headers
            [('Content-Type', 'text/plain'), ('Cache-Control', 'no-cache')]
        """
        return self.replace(headers=(self.headers + [(make_header_name(name), value)]))

    def add_headers(self, headers=[], **kwheaders):
        """
        Return a new response object with the given additional headers.

        Synopsis::

            >>> r = Response(content_type='text/plain')
            >>> r.headers
            [('Content-Type', 'text/plain')]
            >>> r.add_headers(
            ...     cache_control='no-cache',
            ... ).headers
            [('Content-Type', 'text/plain'), ('Cache-Control', 'no-cache')]
        """
        return self.replace(headers=make_headers(self.headers + headers, kwheaders))

    def remove_headers(self, *headers):
        """
        Return a new response object with the named headers removed.

        Synopsis::

            >>> r = Response(content_type='text/plain',
            ...              cache_control='no-cache')
            >>> r.headers
            [('Cache-Control', 'no-cache'), ('Content-Type', 'text/plain')]
            >>> r.remove_headers('Cache-Control').headers
            [('Content-Type', 'text/plain')]
        """
        toremove = {item.lower() for item in headers}
        return self.replace(
            headers=[h for h in self.headers if h[0].lower() not in toremove]
        )

    def add_cookie(
        self,
        name,
        value,
        max_age=None,
        expires=None,
        path: str = "/",
        secure: bool = False,
        domain: t.Optional[str] = None,
        comment: t.Optional[str] = None,
        httponly: bool = False,
        samesite: str = "Lax",
    ):
        """
        Return a new response object with the given cookie added.

        Synopsis::

            >>> r = Response(content_type='text/plain')
            >>> r.headers
            [('Content-Type', 'text/plain')]
            >>> r.add_cookie('a', '1').headers
            [('Content-Type', 'text/plain'), ('Set-Cookie', 'a=1;Version=1')]
        """
        return self.add_header(
            "Set-Cookie",
            Cookie(
                name,
                value,
                max_age,
                expires,
                path,
                secure,
                domain,
                comment=comment,
                httponly=httponly,
                samesite=samesite,
            ),
        )

    def delete_cookie(self, name, path="/", domain=None):
        return self.add_cookie(
            name,
            "",
            max_age=0,
            path=path,
            domain=domain,
            expires=datetime(1970, 1, 1),
        )

    def add_vary(self, *vary_on):
        """
        Return a new response object with the given Vary header.
        Values specified will be added to any existing vary header.

        Synopsis::

            >>> r = Response().add_vary('Accept-Encoding')
            >>> r.headers
            [('Vary', 'Accept-Encoding')]
        """
        _vary_on: Set[str] = set(vary_on)
        newheaders = []
        for k, v in self.headers:
            if k.lower() == "vary":
                _vary_on.update(s.strip() for s in v.split(","))
            else:
                newheaders.append((k, v))
        return self.replace(headers=newheaders + [("Vary", ", ".join(_vary_on))])

    def replace(
        self, content=None, status=None, headers=None, **kwheaders
    ) -> "Response":
        """
        Return a new response object with any of content, status or headers
        changed.

        Synopsis::

            >>> r = Response(content_type='text/html')
            >>> r = r.replace(content='foo',
            ...               status=404,
            ...               headers=[('Content-Type', 'text/plain')],
            ...               content_length=3)

        """
        if content is None:
            content = self.content
            content_iterator = self.content_iterator
            onclose = self.onclose
        else:
            content_iterator = None
            close = getattr(self.content, "close", None)
            onclose = self.onclose
            if close:
                onclose = [close] + onclose

        if headers is None:
            headers = self.headers

        if status is None:
            status = self.status

        if kwheaders:
            toremove = {make_header_name(k) for k in kwheaders}
            headers = [(k, v) for k, v in headers if k not in toremove] + make_headers(
                [], kwheaders
            )

        return self.__class__(
            content,
            status,
            headers,
            onclose=onclose,
            content_iterator=content_iterator,
        )

    def buffered(self):
        """
        Return a new response object with the content buffered into a list.
        This will also generate a content-length header.

        Example usage::

            >>> def generate_content():
            ...     yield "one two "
            ...     yield "three four five"
            ...
            >>> r = Response(content=generate_content())
            >>> r.content  # doctest: +ELLIPSIS
            <generator object ...>
            >>> r = Response(content=generate_content()).buffered()
            >>> r.content
            ['one two ', 'three four five']
        """
        content = list(self.content_iterator)
        content_length = sum(map(len, content))
        return self.replace(content=content, content_length=content_length)

    @property
    def charset(  # type: ignore
        self,
        _parser=re.compile(r";\s*charset=([\w\d\-]+)", re.IGNORECASE).search,
    ) -> str:
        for key, val in self.headers:
            if key == "Content-Type":
                mo = _parser(val)
                if mo:
                    return mo.group(1)
                break
        return default_charset

    @classmethod
    def not_found(cls, request=None) -> "Response":
        """
        Return an HTTP not found response (404).

        Synopsis::

            >>> def view():
            ...     return Response.not_found()
            ...
        """
        return cls(
            status=STATUS_NOT_FOUND,
            content=[
                "<html>\n"
                "<body>\n"
                "    <h1>Not found</h1>\n"
                "    <p>The requested resource could not be found.</p>\n"
                "</body>\n"
                "</html>"
            ],
        )

    @classmethod
    def error(cls, message="500 Internal Server Error") -> "Response":
        """
        Return an HTTP server error response (500).

        Synopsis::

            >>> def view():
            ...     return Response.error()
            ...
        """
        return cls(
            status=STATUS_INTERNAL_SERVER_ERROR,
            content=[
                "<html>\n"
                "<body>\n"
                f"    <h1>{message}</h1>\n"
                "</body>\n"
                "</html>\n"
            ],
        )

    @classmethod
    def unauthorized(cls, authenticate) -> "Response":
        """
        Return an HTTP unauthorized response (401)
        """
        return cls(
            "<html>\n"
            "<body>\n"
            "<h1>401 Authorization Required</h1>\n"
            "</body>\n"
            "</html>",
            status=STATUS_UNAUTHORIZED,
            www_authenticate=authenticate,
        )

    @classmethod
    def unauthorized_basic(cls, realm) -> "Response":
        """
        Return an HTTP unauthorized response (401) with a WWW-Authenticate
        header set for HTTP Basic authentication..
        """
        return cls.unauthorized(authenticate=f'Basic realm="{realm}"')

    @classmethod
    def forbidden(cls, message="Sorry, access is denied") -> "Response":
        """
        Return an HTTP forbidden response (403).

        Synopsis::

            >>> def view():
            ...     return Response.forbidden()
            ...
        """
        return cls(
            "<html>\n<body>\n<h1>%s</h1>\n</body>\n</html>" % (message,),
            status=STATUS_FORBIDDEN,
        )

    @classmethod
    def bad_request(
        cls, message="The server could not understand your request", request=None
    ) -> "Response":
        """
        Return an HTTP bad request response.

        Synopsis::

            >>> def view():
            ...     return Response.bad_request()
            ...

        """
        return cls(
            status=STATUS_BAD_REQUEST,
            content=[
                f"<html>"
                f"<body>"
                f"<h1>Bad request</h1>"
                f"<p>{message}</p>"
                f"</body>"
                f"</html>"
            ],
        )

    @classmethod
    def length_required(cls, request=None) -> "Response":
        """
        Return an HTTP Length Required response (411).

        Synopsis::

            >>> def view():
            ...     return Response.length_required()
            ...

        """
        return cls(
            status=STATUS_LENGTH_REQUIRED,
            content=[
                "<html>"
                "<body>"
                "<h1>A Content-Length header is required</h1>"
                "</body>"
                "</html>"
            ],
        )

    @classmethod
    def payload_too_large(cls, request=None) -> "Response":
        """
        Return an HTTP Payload Too Large response (413)::

            >>> response = Response.payload_too_large()

        """
        return cls(
            status=STATUS_PAYLOAD_TOO_LARGE,
            content=["<html><body><h1>Payload Too Large</h1></body></html>"],
        )

    request_entity_too_large = payload_too_large

    @classmethod
    def method_not_allowed(cls, valid_methods) -> "Response":
        """
        Return an HTTP method not allowed response (405)::

            >>> from fresco import context
            >>> def view():
            ...     if context.request.method == 'POST':
            ...         return Response.method_not_allowed(('POST', ))
            ...

        :param valid_methods: A list of HTTP methods valid for requested URL

        :return: A :class:`fresco.response.Response` instance
        """

        return cls(
            status=STATUS_METHOD_NOT_ALLOWED,
            allow=",".join(valid_methods),
            content=["<html><body><h1>Method not allowed</h1></body></html>"],
        )

    @classmethod
    def internal_server_error(cls) -> "Response":
        """
        Return an HTTP internal server error response (500).

        Synopsis::

            >>> def view():
            ...     return Response.internal_server_error()
            ...

        :return: A :class:`fresco.response.Response` instance
        """

        return cls(
            status=STATUS_INTERNAL_SERVER_ERROR,
            content=["<html><body><h1>Internal Server Error</h1></body></html>"],
        )

    @classmethod
    def unrestricted_redirect(
        cls, location, request=None, status=STATUS_FOUND, **kwargs
    ) -> "Response":
        """
        Return an HTTP redirect reponse (30x).

        :param location: The redirect location or a view specification.
        :param status: HTTP status code for the redirect, default is
                       ``STATUS_FOUND`` (temporary redirect)
        :param kwargs: kwargs to be passed to :func:`fresco.core.urlfor` to
                       construct the redirect URL

        Synopsis:

            >>> def view():
            ...   return Response.redirect("/new-location")
            ...

        The location parameter is interpreted as follows:

        - If it is a callable it is assumed to be a view,
          and passed to :func:`fresco.core.urlfor`, along with any keyword
          arguments, to generate the redirect URL.

        - If it is a string and contains '://' it is assumed to be an absolute
          URL and no further processing is done.

        - If it is a string and contains a slash,
          it is assumed to be a
          relative URL and resolved relative to the current request.

        - If it is any other string
          it is passed to :func:`fresco.core.urlfor` to resolve it to a URL.
          If this fails (ie raises RouteNotFound) it is assumed to be a
          relative URL and resolved relative to the current request.
        """
        if callable(location):
            location = fresco.context.app.urlfor(location, **kwargs)

        elif "://" in location:
            pass

        elif "/" in location:
            if request is None:
                request = fresco.context.request
            location = request.resolve_url(location)

        else:
            from fresco.routing import RouteNotFound

            try:
                location = fresco.context.app.urlfor(location, **kwargs)
            except RouteNotFound:
                if kwargs:
                    # kwargs only makes sense if this is a view spec
                    raise
                if request is None:
                    request = fresco.context.request
                location = request.resolve_url(location)

        return cls(
            "<html><head></head><body>\n"
            "<h1>Page has moved</h1>\n"
            "<p><a href='%s'>%s</a></p>\n"
            "</body></html>" % (location, location),
            status=status,
            location=location,
        )

    @classmethod
    def unrestricted_redirect_permanent(cls, *args, **kwargs) -> "Response":
        """
        Return an HTTP permanent redirect reponse.

        :param location: the URI of the new location. If relative this will be
                         converted to an absolute URL based on the current
                         request.

        """
        kwargs["status"] = STATUS_MOVED_PERMANENTLY
        return cls.unrestricted_redirect(*args, **kwargs)

    @classmethod
    def unrestricted_redirect_temporary(cls, *args, **kwargs) -> "Response":
        """
        Return an HTTP permanent redirect reponse.

        :param location: the URI of the new location. If relative this will be
                         converted to an absolute URL based on the current
                         request.

        """
        kwargs["status"] = STATUS_FOUND
        return cls.unrestricted_redirect(*args, **kwargs)

    @classmethod
    def redirect(
        cls,
        url,
        fallback=None,
        status=STATUS_FOUND,
        _is_safe_url=is_safe_url,
        allowed_hosts=frozenset(),
        **kwargs,
    ) -> "Response":
        """
        Return an HTTP redirect reponse (30x). Will only redirect to the
        current host or hosts in the ``allowed_hosts`` list. A ``ValueError``
        will be raised if the URL is not permitted and no fallback is
        specified.

        :param location: The redirect location or a view specification.
        :param status: HTTP status code for the redirect, default is
                       ``STATUS_FOUND`` (temporary redirect)
        :param fallback: a fallback URL to be used for the redirect in the case
                         that ``location`` is considered unsafe
        :param kwargs: kwargs to be passed to :func:`fresco.core.urlfor` to
                       construct the redirect URL, or the fallback in the case
                       that ``location`` is already a qualified URL.

        Synopsis::

            >>> def view():
            ...   return Response.redirect("/new-location")
            ...

        The location argument is interpreted as for
        :meth:`~fresco.response.Response.unrestricted_redirect`
        """
        # Create the fallback redirect response always so that
        # changes to route names don't result in latent bugs
        fallback_response = None
        if fallback:
            fallback_response = cls.unrestricted_redirect(
                fallback, status=status, **kwargs
            )
        if callable(url) or "//" not in url:
            return cls.unrestricted_redirect(url, status=status, **kwargs)

        if url and _is_safe_url(url, allowed_hosts):
            return cls.unrestricted_redirect(url, status=status, **kwargs)
        if fallback_response:
            return fallback_response
        raise ValueError("Unsafe URL")

    @classmethod
    def redirect_permanent(cls, *args, **kwargs) -> "Response":
        """
        Return an HTTP permanent redirect reponse.

        :param location: the URI of the new location. If relative this will be
                         converted to an absolute URL based on the current
                         request.

        """
        kwargs["status"] = STATUS_MOVED_PERMANENTLY
        return cls.redirect(*args, **kwargs)

    @classmethod
    def redirect_temporary(cls, *args, **kwargs) -> "Response":
        """
        Return an HTTP permanent redirect reponse.

        :param location: the URI of the new location. If relative this will be
                         converted to an absolute URL based on the current
                         request.

        """
        kwargs["status"] = STATUS_FOUND
        return cls.redirect(*args, **kwargs)

    @classmethod
    def meta_refresh(cls, location, delay=1, request=None) -> "Response":
        """
        Return an HTML page containing a <meta http-equiv="refresh"> tag,
        causing the browser to redirect to the given location after ``delay``
        seconds.

        :param location: the URI of the new location. If relative this will be
                         converted to an absolute URL based on the current
                         request.

        """
        if "://" not in location:
            if request is None:
                request = fresco.context.request
            location = request.resolve_url(location)
        return cls(
            [
                (
                    "<!DOCTYPE html>"
                    "<html>"
                    '<head><meta http-equiv="refresh" content="0; url={0}"></head>'
                    '<body><p><a href="{0}">Click here to continue</a></p></body>'
                    "</html>"
                ).format(location)
            ],
            content_type="text/html",
        )

    @classmethod
    def json(
        cls,
        data,
        indent=None,
        separators=(",", ":"),
        content_type="application/json",
        status=None,
        headers=None,
        dumps=stdlib_json.dumps,
        **kwargs,
    ) -> "Response":
        """
        Return an ``application/json`` response with the given data
        JSON serialized

        :param data: The data to json encode.
        :param indent: The indent level.
                       Defaults to ``None`` (no pretty printing)
        :param separators: Defaults to ``(',', ':')`` for the most compact JSON
                           representation
        :param kwargs: Other keyword arguments are passed to ``json.dumps``.
                       These may be used to change encoding paramters, for
                       example overriding the default ``JSONEncoder`` class.

        """
        return cls(
            [dumps(data, indent=indent, separators=separators, **kwargs)],
            status=status,
            headers=headers,
            content_type=content_type,
        )


def dump_response(r, line_break=b"\r\n", encoding="UTF-8"):
    """
    Return a byte-string representation of the given response, as for a HTTP
    response.
    """
    output = []
    output.append(r.status.encode("ascii"))
    output.append(line_break)
    for k, v in sorted(r.headers):
        output.append(k.encode("ascii"))
        output.append(b": ")
        output.append(v.encode("ascii"))
        output.append(line_break)
    output.append(line_break)
    output.extend(r.content_iterator)
    s = b"".join(output)
    if encoding:
        return s.decode(encoding)
    return s



from itertools import chain
from typing import Any

from fresco.core import context
from fresco.request import Request
from fresco.response import Response
from fresco.routing import Route
from fresco.routing import GET
from fresco.routing import RouteArg
from fresco.util.http import parse_header
from fresco.util.wsgi import make_environ

__all__ = ["subrequest", "subrequest_raw", "subrequest_bytes"]


class Markup(str):
    def __html__(self):
        return self


def subrequest(view, *args, **kwargs):
    """
    Return the response generated by a subrequest to ``view``

    This behaves differently depending on the value of ``view``:

    - **A string containing a ``/``**:
      a full request context is constructed to emulate a request
      to the given path.
      The subrequest will invoke all middleware layers and application hooks.

    - **A callable**: a new subrequest context will be created and
      ``view(*args, **kwargs)`` will be called. Middleware and application
      hooks will not be called, and any
      :class:`~fresco.routeargs.RouteArg` instances will not be resolved.

    - **Any other value**: the view function is looked up using the same
      route resolution rules as :meth:`~fresco.core.FrescoApp.urlfor`.
      Middleware and hooks will be skipped, but
      :class:`~fresco.routeargs.RouteArg` instances will be resolved.

    If you pass in a view callable you can force RouteArgs to be resolved by
    specifying ``_resolve=True``.

    You can force a full request to be emulated by specifying ``_full=True``.

    Passing additional positional and keyword arguments has a different
    effect depending on the value of ``view``, ``_resolve`` and ``_full``:

    - **If view is a path, or if _full is True**,
      ``*args`` and ``**kwargs`` are passed to
      :meth:`~fresco.core.FrescoApp.requestcontext`,
      and may be used to customize the WSGI environment
      used for the subrequest.
      A new request context will be created
      from a copy of the current request's WSGI environ.
      This will contain the same values
      except for ``PATH_INFO`` and ``QUERY_STRING``.

    - **If view is a callable and _resolve is False**,
      ``*args`` and ``**kwargs`` are passed to the view callable.
      No new request context is created.

    - **Otherwise**
      ``view`` is assumed to be the name of a route.
      ``*args`` and ``**kwargs`` are passed to the route lookup mechanism
      and should provide the values defined in the route's path.
      ``kwargs`` values also override any :class:`~fresco.routeargs.RouteArg`s
      defined by the route. These will only be generated from the request if
      a corresponding value is not present in ``kwargs``.
      No new request context is created.

    :param view: The target path, view callable, or viewspec.

    :param _mode: One of ``raw``, ``str`` or ``bytes``.
                 If ``str``, the content of the subrequest will be
                 returned as a string.
                 If ``bytes``, the content will be returned as a
                 byte string.
                 If ``raw``, the raw response will be returned. In
                 this case the caller is responsible for iterating the
                 content and calling any functions in ``response.onclose``.

    :param _resolve: if ``True``, callable views will be resolved via the route
                     traversal mechanism, allowing routed arguments to be
                     set from the current request.

    :param _full: if ``True``, a full new request context will be set up
                  ensuring application hooks and middleware are called
                  for the subrequest.
                  The WSGI environ will be a copy of the current request's environ
                  with only ``PATH_INFO`` and ``QUERY_STRING`` set up for the
                  specified view.
                  Additional arguments
                  are passed to :meth:`~fresco.core.FrescoApp.requestcontext`
                  and can be used to override values in the new request environ.

    :param _env: Additional keys to merge into the subrequest's environ
                 (may only be used when ``_full`` is True)

    :param args: Additional positional arguments

    :param kwargs: Additional keyword arguments
    """
    mode = kwargs.pop("_mode", "str")
    if mode == "str":
        return subrequest_str(view, *args, **kwargs)
    elif mode == "bytes":
        return subrequest_bytes(view, *args, **kwargs)
    elif mode == "raw":
        return subrequest_raw(view, *args, **kwargs)
    else:
        raise ValueError("Mode must be one of 'str', 'bytes', or 'raw'")


def subrequest_str(view, *args, **kwargs) -> str:
    """
    Perform a subrequest and return the response content as a string.
    See :func:`~fresco.subrequests.subrequest` for details
    """
    return response_to_str(subrequest_raw(view, *args, **kwargs))


def subrequest_bytes(view, *args, **kwargs) -> bytes:
    """
    Perform a subrequest and return the response content as a byte string.
    See :func:`~fresco.subrequests.subrequest` for details
    """
    return response_to_bytes(subrequest_raw(view, *args, **kwargs))


def subrequest_raw(view, *args, **kwargs) -> Response:
    """
    Perform a subrequest and return the raw response object.
    See :func:`~fresco.subrequests.subrequest` for details
    """

    current_request = context.request
    response = None
    full = kwargs.pop("_full", False)
    resolve = kwargs.pop("_resolve", False)
    _env = kwargs.pop("_env", {})
    environ = current_request.environ.copy()
    del environ["PATH_INFO"]
    if "QUERY_STRING" in environ:
        del environ["QUERY_STRING"]

    app = context.app
    context.push(**dict(context.currentcontext(), is_subrequest=True))
    try:
        if full or (isinstance(view, str) and "/" in view):
            del environ[current_request.STATE_ENV_KEY]
            if full:
                path = app.urlfor(view, *args, **kwargs)
            else:
                path = view
            environ = make_environ(path, environ, b"")
            if _env:
                environ.update(_env)
            request = app.request_class(environ)
            context["request"] = request
            response = app.view(request)
        else:
            assert not _env, "Use of _env invalid without _full"
            if not callable(view) or resolve:
                view, routed_args, routed_kwargs = resolve_viewspec(
                    view, *args, **kwargs
                )

                # Use args/kwargs extracted during route traversal,
                args, kwargs = routed_args, routed_kwargs

            response = view(*args, **kwargs)

        return response  # type: ignore
    finally:
        context.pop()


def response_to_str(response) -> str:
    ct, params = parse_header(response.get_header("Content-Type"))
    encoding = params.get("charset")
    try:
        # Check for the common case that response.content is already a
        # string
        if isinstance(response.content, str):
            content = response.content

        # ...or a list of strings?
        elif isinstance(response.content, list) and all(
            isinstance(i, str) for i in response.content
        ):
            content = "".join(response.content)

        # Anything else: decode the content iterator
        else:
            if not encoding:
                raise ValueError("No content encoding specified")
            content = b"".join(response.content_iterator).decode(encoding)
    finally:
        for f in response.onclose:
            f()
    if ct.startswith("text/html") or "xml" in ct.split(";")[0]:
        return Markup(content)
    return content


def response_to_bytes(response):
    try:
        return b"".join(response.content_iterator)
    finally:
        for f in response.onclose:
            f()


def resolve_viewspec(viewspec, *args, **kwargs):
    """
    :param viewspec: a view name, a reference in the form
                     ``'package.module.viewfunction'``, or the view
                     callable itself.
    :param args: positional arguments for the route path traversal.
    :param kwargs: keyword arguments for the route path traversal.
    :return: A tuple of ``(<view callable>, args, kwargs)``.
             ```args`` and ``kwargs`` are the arguments to apply
             to ``view`` as extracted during the route traversal.
    """
    app = kwargs.pop("_routecollection", context.app)
    request = kwargs.pop("request", None) or context.request
    method = kwargs.pop("method", GET)

    if isinstance(viewspec, str) and ":" in viewspec:
        viewspec, remainder = viewspec.split(":", 1)
        delegated_route = app.routefor(viewspec)

        for k in delegated_route.routed_args_default:
            if k not in kwargs:
                v = delegated_route.routed_args_default[k]
                if callable(v):
                    v = v(request)
                kwargs[k] = v

        delegate_args, delegate_kwargs = _get_args_for_route(
            delegated_route, request, args, kwargs
        )

        rc = delegated_route.routecollectionfactory(*delegate_args, **delegate_kwargs)
        return resolve_viewspec(
            remainder,
            _routecollection=rc,
            request=request,
            method=method,
            *args,
            **kwargs
        )

    route = app.routefor(viewspec)
    view = route.getview(method)
    view_args, view_kwargs = _get_args_for_route(route, request, args, kwargs)
    return view, view_args, view_kwargs


def _get_args_for_route(
    route: Route, request: Request, args: tuple[Any, ...], kwargs: dict[str, Any]
) -> tuple[tuple[Any, ...], dict[str, Any]]:
    """
    Return the args/kwargs required to pass to the view callable for ``route``.
    """
    mutable_args = list(args)
    mutable_kwargs = kwargs.copy()
    route_kwargs = {}
    route_args = list(
        chain(
            [request] if route.provide_request else [],
            ((a(request) if isinstance(a, RouteArg) else a) for a in route.view_args),
            (
                mutable_args.pop(0)
                for a in route.pattern.segments
                if a.converter and not a.name
            ),
        )
    )

    required_kwargs = chain(
        route.view_kwargs.keys(),
        (s.name for s in route.pattern.segments if s.converter and s.name),
    )

    _marker = object()
    for k in required_kwargs:
        v = mutable_kwargs.pop(k, _marker)
        if v is not _marker:
            route_kwargs[k] = v
            continue

        v = route.routed_args_default.get(k, _marker)
        if v is not _marker:
            if callable(v):
                v = v(request)
            route_kwargs[k] = v
            continue

        v = route.view_kwargs.get(k, _marker)
        if v is not _marker:
            v = v(request) if isinstance(v, RouteArg) else v
            route_kwargs[k] = v

    return tuple(route_args), route_kwargs



# encoding=UTF-8
# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
from fresco import Route
from fresco.routing import GET
from fresco.response import Response

# A unicode path that can't be represented using ASCII encoding
unquoted_unicode_path = "/ø"

# IRI quoted version of the same path. This is the string the server
# would receive as the HTTP Request-URI
quoted_unicode_path = "/%C3%B8"

# WSGI encoded version of the path. This is the string that appears in the
# environ dict.
wsgi_unicode_path = b"/\xc3\xb8".decode("latin1")

# Malformed path, as sent to the server by a non conforming client.
# The code point has been incorrectly encoded in latin-1 instead of
# UTF-8
misquoted_wsgi_unicode_path = b"/\xf8".decode("latin1")


class CBV(object):
    """
    A class based view
    """

    __routes__ = [
        Route("/", GET, "index_html"),
        Route("/page", GET, "view_page"),
        Route("/page2", GET, "view_page", tag="page2"),
    ]

    def __init__(self, s):
        self.s = s

    def index_html(self):
        return Response([self.s])

    def view_page(self):
        return Response([])


def module_level_function():
    """
    A module level function
    """


def wsgi_app(environ, start_response):
    start_response("200 OK", [("Content-Type", "text/plain")])

    return [b"ok"]



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
import typing as t

from fresco.types import WSGIApplication
from fresco.types import WSGIEnviron
from fresco.types import StartResponse

__all__ = ["XForwarded"]


class XForwarded(object):
    """
    Modify the WSGI environment so that the X_FORWARDED_* headers are observed
    and generated URIs are correct in a proxied environment.

    Use this whenever the WSGI application server is sitting behind
    Apache or another proxy server.

    It is easy for clients to spoof the X-Forwarded-For header. You can largely
    protect against this by listing all trusted proxy server addresses in
    ``trusted``. See http://en.wikipedia.org/wiki/X-Forwarded-For for more
    details.

    HTTP_X_FORWARDED_FOR is substituted for REMOTE_ADDR and
    HTTP_X_FORWARDED_HOST for SERVER_NAME. If HTTP_X_FORWARDED_SSL is set, then
    the wsgi.url_scheme is modified to ``https`` and ``HTTPS`` is set to
    ``on``.

    :param trusted:
        List of IP addresses trusted to set the HTTP_X_FORWARDED_* headers

    :param force_https:
        If True, the following environ keys will be set unconditionally:
        ``"HTTPS": "on"`` and ``"wsgi.url_scheme": "https"`` will be set

    Example::

        >>> from fresco import FrescoApp, context, GET, Response
        >>> from flea import Agent
        >>> app = FrescoApp()
        >>> @app.route('/', GET)
        ... def view():
        ...     return Response(["URL is ", context.request.url,
        ...                      "; REMOTE_ADDR is ",
        ...                      context.request.remote_addr])
        ...
        >>> r = Agent(XForwarded(app))
        >>> response = r.get('/',
        ...     SERVER_NAME='wsgiserver-name',
        ...     SERVER_PORT='8080',
        ...     HTTP_HOST='wsgiserver-name:8080',
        ...     REMOTE_ADDR='127.0.0.1',
        ...     HTTP_X_FORWARDED_HOST='real-name:81',
        ...     HTTP_X_FORWARDED_FOR='1.2.3.4'
        ... )
        >>> response.body
        u'URL is http://real-name:81/; REMOTE_ADDR is 1.2.3.4'
        >>> response = r.get('/',
        ...     SERVER_NAME='wsgiserver-name',
        ...     SERVER_PORT='8080',
        ...     HTTP_HOST='wsgiserver-name:8080',
        ...     REMOTE_ADDR='127.0.0.1',
        ...     HTTP_X_FORWARDED_HOST='real-name:443',
        ...     HTTP_X_FORWARDED_FOR='1.2.3.4',
        ...     HTTP_X_FORWARDED_SSL='on'
        ... )
        >>> response.body
        u'URL is https://real-name/; REMOTE_ADDR is 1.2.3.4'
    """

    def __init__(
        self,
        app: WSGIApplication,
        trusted: t.Optional[t.Iterable[str]] = None,
        force_https: t.Optional[bool] = None,
    ) -> None:
        self.app = app
        self.force_https = force_https
        if trusted:
            self.trusted = set(trusted)
        else:
            self.trusted = set()

    def __call__(
        self, environ: WSGIEnviron, start_response: StartResponse
    ) -> t.Iterable[bytes]:
        """
        Call the WSGI app, passing it a modified environ
        """
        env = environ.get

        if self.force_https is None:
            is_ssl = (
                env("HTTP_X_FORWARDED_PROTO") == "https"
                or env("HTTP_X_FORWARDED_SSL") == "on"
            )
        else:
            is_ssl = self.force_https
        host = env("HTTP_X_FORWARDED_HOST")

        if host is not None:
            if ":" in host:
                port = host.split(":")[1]
            else:
                port = is_ssl and "443" or "80"

            environ["HTTP_HOST"] = host
            environ["SERVER_PORT"] = port

        if is_ssl:
            environ["wsgi.url_scheme"] = "https"
            environ["HTTPS"] = "on"

        forwarded_for = env("HTTP_X_FORWARDED_FOR")
        if forwarded_for:
            addrs = forwarded_for.split(", ")

            if self.trusted:
                for ip in addrs[::-1]:
                    # Find the first non-trusted ip; this is our remote address
                    if ip not in self.trusted:
                        environ["REMOTE_ADDR"] = ip
                        break
            else:
                environ["REMOTE_ADDR"] = addrs[0]

        return self.app(environ, start_response)



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
#
"""
Utilities for wrapping IO streams. These are used internally by
fresco when parsing wsgi request input streams.
"""
from io import DEFAULT_BUFFER_SIZE

import typing as t


ByteIterator = t.Iterator[bytes]


def io_iterator(fp: t.IO, size=DEFAULT_BUFFER_SIZE, maxlen=-1) -> ByteIterator:
    bytecount = 0
    if maxlen > 0:
        while True:
            r = fp.read(min(maxlen - bytecount, size))
            bytecount += len(r)
            if r == b"":
                return
            yield r
            if bytecount >= maxlen:
                break
    else:
        while True:
            r = fp.read(size)
            if r == b"":
                return
            yield r
